--[[
 Companion module for MiniPet
 Version 5.1.248
 $Revision: 183 $

 All credit for authoring this addon is listed below
]]--

--[[
Copyright (c) 2010-2012, LordFarlander
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
]]--

local L = LibStub( "AceLocale-3.0" ):GetLocale( "MiniPet" )

local LibLordFarlander = LibStub( "LibLordFarlander-2.0", true )
if( not LibLordFarlander ) then
    return
end--if

local ItemList = LibStub( "LibLordFarlander-ItemList-2.0", true )
if( not ItemList ) then
    return
end--if

local PetAndMountDatabase = LibStub( "LibLordFarlander-PetAndMountDatabase-1.1", true )
if( not PetAndMountDatabase ) then
    return
end--if

if( not LibStub( "LibLordFarlander-Randomizer-2.0", true ) ) then
    return
end--if

local TR = LibStub( "LibLordFarlander-TableRecycler-1.0", true )
if( not TR ) then
    return
end--if

local moduleName = "Companions"

local Companions = {
    displayname = L["Companions"],
    events = {
        "AddBagItem",
        "AddInventoryItem",
		"BuildAdditionalOptions",
        "BuildZoneFavorites",
        "CheckItem",
        "ChooseRandom",
        "DismissCompanion",
        "FinalizeBag",
        "GetAllCompanionsList",
        "GetCurrentCompanion",
        "GetZoneList",
        "HasCompanions",
        "HasDismissableCompanionOut",
        "IsCompanionSummonInCooldown",
        "MouseClick",
        "OutputWCPCode",
        "ScanPlayer",
        "SetCurrentCompanion",
        "StartBagScan",
        "SummonCompanion",
        "COMPANION_UPDATE",
        "PLAYER_REGEN_DISABLED",
        "PLAYER_REGEN_ENABLED",
        "ZONE_CHANGED_NEW_AREA",
    },
    settings = {
        [L["boostNew"]] = {
            type = "toggle",
            order = 1,
            name = L["Boost new companions"],
            width = "double",
            arg = {
                DescriptionString = L["Toggles if new companions will be selected more often than old ones."],
                VariableName = "BoostNew",
                Default = false,
            },
            desc = MiniPet.BooleanOptionDescription,
            get = "GetGenericOption",
            set = "SetOption_ResetPets",
        },
        [L["boostNewTime"]] = {
            type = "range",
            order = 2,
            name = L["New time"],
            cmdHidden = true,
            min = 0.0,
            max = 31.0,
            bigStep = 1.0,
            width = "double",
            desc = L["Set for how many days companions are concidered new."],
            get = function( info )
                      return info.handler:GetSetting( moduleName, "profile", "BoostNewTime" ) / 86400
                  end,
            set = function( info, val )
                      info.handler:SetSetting( moduleName, "profile", "BoostNewTime", val * 86400 )
                  end,
        },
        [L["list"]] = {
            type = "execute",
            order = 3,
            cmdHidden = true,
            name = L["List all companions"],
            desc = L["Show a list with all of your companions."],
            func = "ListCompanions",
        },
        [L["refresh"]] = {
            type = "execute",
            order = 4,
            cmdHidden = true,
            width = "double",
            name = L["Refresh companion list"],
            desc = L["Force a refresh of the list of all your companions."],
            func = "ResetCompanions",
        },
        [L["wcpCode"]] = {
            type = "execute",
            order = 5,
            width = "double",
            name = L["Get WarcraftPets.com code"],
            desc = L["Outputs a code to the chat log you can use to update your WarcraftPets.com collection."],
            func = function( info )
                       info.handler:DoCustomEvent( "OutputWCPCode" )
                   end,
        },
    },
    required = true,
    -- Module variables
    currentCritterIndex  = nil,
    currentCompanion     = nil,
    playersCompanions    = { zones = { default = TR:GetTable(), }, list = {}, },
    dirty                = false,
    dirtyBags            = { [-254] = true, [-255] = true, },
}

for bag = 0, NUM_BAG_SLOTS do
    Companions.dirtyBags[bag] = true
end--for

function Companions:BuildFavoritesOptions( MiniPet )
    local favoritesOptions = {
        type = "group",
        name = L["Favorites"],
        desc = L["Companions that should be picked more often if you have them."],
        handler = MiniPet,
        args = {
            favorite = {
                type = "select",
                order = 1,
                name = L["Favorites"],
                desc = L["Companions that should be picked more often if you have them."],
                width = "full",
                cmdHidden = true,
                values = function( info )
							 local handler = info.handler
							 local ui = handler:GetUISettings( moduleName )

                             ui.favorites = ui.favorites and table.wipe( ui.favorites ) or {}

                             local favorites = ui.favorites
							 local profileFavorites = handler:GetSetting( moduleName, "profile", "Favorites" )
                             local firstItem = 0
                             local used = TR:GetTable()
                             local companions = handler:GetAllCompanionsList()

                             for _, pet in pairs( companions ) do
                                 local id = pet:GetID()

                                 used[id] = true
                                 tinsert( favorites, id )
                                 if( type( profileFavorites[id] ) == "boolean" ) then
                                     profileFavorites[id] = 2.0
                                 end--if
                                 if( ( type( profileFavorites[id] ) ~= "nil" ) and ( profileFavorites[id] == 1.0 ) ) then
                                     profileFavorites[id] = nil
                                 end--if
                             end--for
                             for pet, howOften in pairs( profileFavorites ) do
                                 if( ( not used[pet] ) and ( ( pet < -1 ) and GetSpellInfo( pet * -1 ) or GetItemInfo( pet ) ) ) then
                                     tinsert( favorites, pet )
                                 end--if
                             end--if
                             TR:ReturnTable( used )
                             table.sort( favorites, function( a, b ) return ( ( a < 0 ) and GetSpellInfo( a * -1 ) or GetItemInfo( a ) ) < ( ( b < 0 ) and GetSpellInfo( b * -1 ) or GetItemInfo( b ) ) end )
                             favorites = TR:GetTable()
                             
                            local lHowOften = L["%.1f times as often"]
                            local lDefault = L["default"]
                            local lNever = L["never"]
                             
                             for index, pet in pairs( ui.favorites ) do
                                 local sName, sTexture
                                 local sHowOften = lDefault
                                 local howOften = profileFavorites[pet] or 1.0

                                 if( pet < 0 ) then
                                     sName, _, sTexture, _, _, _, _, _, _ = GetSpellInfo( pet * -1 )
                                 else
                                     sName, _, _, _, _, _, _, _, _, sTexture = GetItemInfo( pet )
                                 end--if
                                 if( howOften == 0.0 ) then
                                     sHowOften = lNever
                                 elseif( howOften ~= 1.0 ) then
                                     sHowOften = lHowOften:format( howOften )
                                 end--if
                                 if( sTexture ) then
                                     favorites[index] = ("|T%s:0|t %s |cFFFF0000[%s]|r"):format( sTexture, sName, sHowOften )
                                 else
                                     favorites[index] = ("%s |cFFFF0000[%s]|r"):format( sName, sHowOften )
                                 end--if
                                 if( firstItem == 0 ) then
                                     firstItem = index
                                 end--if
                             end--for
                             if( ( not ui.currentFavorite ) or ( not favorites[ui.currentFavorite] ) ) then
                                 ui.currentFavorite = firstItem
                             end--if
                             TR:ReturnTable( table.wipe( companions ) )
                             return favorites
                         end,
                get = function( info )
                          return info.handler:GetUISettings( moduleName ).currentFavorite
                      end,
                set = function( info, val )
                          info.handler:GetUISettings( moduleName ).currentFavorite = val
                      end,
            },
            frequency = {
                type = "range",
                order = 2,
                name = L["Frequency"],
                cmdHidden = true,
                min = 0.0,
                max = 11.0,
                bigStep = 0.1,
                isPercent = true,
                width = "double",
                desc = L["Set how often the companion gets picked.  2 is twice as often and 0 is never."],
                get = function( info )
                          local ui = info.handler:GetUISettings( moduleName )

                          return info.handler:GetSetting( moduleName, "profile", "Favorites")[ui.favorites[ui.currentFavorite]] or 1.0
                      end,
                set = function( info, val )
                          local ui = info.handler:GetUISettings( moduleName )

                          if( val == 1.0 ) then
                              info.handler:GetSetting( moduleName, "profile", "Favorites" )[ui.favorites[ui.currentFavorite]] = nil
                          else
                              info.handler:GetSetting( moduleName, "profile", "Favorites" )[ui.favorites[ui.currentFavorite]] = val
                          end--if
                          info.handler:BuildZoneFavorites()
                      end,
            },
            [L["list"]] = {
                type = "execute",
                order = 4,
                guiHidden = true,
                name = L["List all favorite companions"],
                desc = L["Show a list with all of your favorite companions."],
                func = function( info )
                           local handler = info.handler
                           local Print = info.handler.Print
                           local lHowOften = L["%.1f times as often"]
                           local lNever = L["never"]
                           local GetItemLink = LibLordFarlander.GetItemLink
                
                           Print( handler, L["List of your favorite companions:"] )
                           for pet, howOften in pairs( handler:GetSetting( moduleName, "profile", "Favorites" ) ) do
                               if( howOften ~= 1.0 ) then
                                   local link = ( pet > 0 ) and GetItemLink( pet ) or GetSpellLink( pet * -1 )

                                   if( pet > 0 ) then
                                       sTexture = select( 10, GetItemInfo( pet ) )
                                   else
                                       sTexture = select( 3, GetSpellInfo( pet * -1 ) )
                                   end--if
                                   if( howOften == 0.0 ) then
                                       howOften = lNever
                                   else
                                       howOften = lHowOften:format( howOften )
                                   end--if

                                   if( sTexture ) then
                                       Print( handler, ("|T%s:0|t %s |cFFFF0000[%s]|r"):format( sTexture, link or tostring( pet ), howOften ) )
                                   else
                                       Print( handler, ("%s |cFFFF0000[%s]|r"):format( link or tostring( pet ), howOften ) )
                                   end--if
                               end--if
                           end--for
                       end,
            },
        },
    }

	return favoritesOptions
end--Companions:BuildFavoritesOptions( MiniPet )

function Companions:BuildZoneFavoritesOptions( MiniPet )
    local zoneTemplate = {
        header = {
            type = "header",
            order = 1,
            width = "full",
            cmdHidden = true,
            name = function( info )
                       return LibLordFarlander.UnnamifyString( info[#info - 1] )
                   end,
        },
        minipet = {
            type = "select",
            order = 2,
            name = L["Companion"],
            desc = L["Companions that should be picked more often if you have them in the selected zone."],
            width = "full",
            cmdHidden = true,
            values = function( info )
                         local ui = info.handler:GetUISettings( moduleName )

                         ui.favorites = ui.favorites and table.wipe( ui.favorites ) or {}

                         local favorites = ui.favorites
                         local used = TR:GetTable()
                         local pZoneFavorites = info.handler:GetSetting( moduleName, "profile", "ZoneFavorites" )
                         local zoneFavorite = pZoneFavorites[LibLordFarlander.UnnamifyString( info[#info - 1] )]
                         local companions = info.handler:GetAllCompanionsList()
                         local pFavorites = info.handler:GetSetting( moduleName, "profile", "Favorites" )

                         for _, pet in pairs( companions ) do
                             local id = pet:GetID()

                             used[id] = true
                             tinsert( favorites, id )
                             if( type( pFavorites[id] ) == "boolean" ) then
                                 pFavorites[id] = 2.0
                             end--if
                             if( ( type( pFavorites[id] ) ~= "nil" ) and ( pFavorites[id] == 1.0 ) ) then
                                 pFavorites[id] = nil
                             end--if
                         end--for
                         if( zoneFavorite and ( type( zoneFavorite ) ~= "table" ) ) then
                             pZoneFavorites[LibLordFarlander.UnnamifyString( info[#info - 1] )] = { [zoneFavorite] = { frequency = 1.0, exclusive = true } }
                             zoneFavorite = pZoneFavorites[LibLordFarlander.UnnamifyString( info[#info - 1] )]
                         end--if
                         if( not zoneFavorite ) then
                             pZoneFavorites[LibLordFarlander.UnnamifyString( info[#info - 1] )] = TR:GetTable()
                             zoneFavorite = pZoneFavorites[LibLordFarlander.UnnamifyString( info[#info - 1] )]
                         end--if
                         if( zoneFavorite ) then
                             for pet, _ in pairs( zoneFavorite ) do
                                 if( not used[pet] and ( ( pet < -1 ) and GetSpellInfo( pet * -1 ) or GetItemInfo( pet ) ) ) then
                                     tinsert( favorites, pet )
                                     used[pet] = true
                                 end--if
                             end--for
                         end--if
                         TR:ReturnTable( used )
                         table.sort( favorites, function( a, b ) return ( ( a < 0 ) and GetSpellInfo( a * -1 ) or GetItemInfo( a ) ) < ( ( b < 0 ) and GetSpellInfo( b * -1 ) or GetItemInfo( b ) ) end )
                         favorites = TR:GetTable()
                         
                         local lExclusive = L["exclusive"]
                         local lDefault = L["default"]
                         
                         for index, pet in pairs( ui.favorites ) do
                             local sName, sTexture
                             local howOften = 1.0
                             local sHowOften = lDefault
                             local exclusive = false

                             if( zoneFavorite[pet] ) then
                                 howOften = zoneFavorite[pet].frequency or 1.0
                                 exclusive = zoneFavorite[pet].exclusive
                             end--if
                             if( pet < 0 ) then
                                 sName, _, sTexture, _, _, _, _, _, _ = GetSpellInfo( pet * -1 )
                             else
                                 sName, _, _, _, _, _, _, _, _, sTexture = GetItemInfo( pet )
                             end--if
                             if( sTexture ) then
                                 favorites[index] = ("|T%s:0|t %s"):format( sTexture, sName )
                             end--if
                             if( howOften == 0.0 ) then
                                 sHowOften = L["never"]
                             elseif( howOften ~= 1.0 ) then
                                 sHowOften = L["%.1f times as often"]:format( howOften )
                             end--if
                             if( exclusive ) then
                                 sHowOften = string.format( "%s, %s", sHowOften, lExclusive )
                             end--if
                             if( sTexture ) then
                                 favorites[index] = ("|T%s:0|t %s |cFFFF0000[%s]|r"):format( sTexture, sName, sHowOften )
                             end--if
                         end--for
                         if( #favorites == 0 ) then
                             ui.currentFavorite = nil
                             favorites[0] = L["none"]
                         elseif( ( currentFavorite == nil ) or ( ui.currentFavorite > #favorites ) ) then
                             ui.currentFavorite = 1
                         end--if
                         TR:ReturnTable( table.wipe( companions ) )
                         return favorites
                     end,
            get = function( info )
                      return info.handler:GetUISettings( moduleName ).currentFavorite
                  end,
            set = function( info, val )
                      info.handler:GetUISettings( moduleName ).currentFavorite = val
                  end,
            },
            frequency = {
                type = "range",
                order = 3,
                name = L["Frequency"],
                cmdHidden = true,
                min = 0.0,
                max = 11.0,
                bigStep = 0.1,
                isPercent = true,
                width = "double",
                desc = L["Set how often the companion gets picked.  2 is twice as often and 0 is never."],
                get = function( info )
                          local ui = info.handler:GetUISettings( moduleName )
                          local zoneFavorite = info.handler:GetSetting( moduleName, "profile", "ZoneFavorites" )[LibLordFarlander.UnnamifyString( info[#info - 1] )]

                          if( zoneFavorite and zoneFavorite[ui.favorites[ui.currentFavorite]] ) then
                              return zoneFavorite[ui.favorites[ui.currentFavorite]].frequency or 1.0
                          end--if
                          return 1.0
                      end,
                set = function( info, val )
                          local zoneFavorite = info.handler:GetSetting( moduleName, "profile", "ZoneFavorites" )[LibLordFarlander.UnnamifyString( info[#info - 1] )]
                          local currentFavorite
                          local ui = info.handler:GetUISettings( moduleName )

                          if( val == 1.0 ) then
                              val = nil
                          end--if
                          if( not zoneFavorite ) then
                              info.handler:GetSetting( moduleName, "profile", "ZoneFavorites" )[LibLordFarlander.UnnamifyString( info[#info - 1] )] = TR:GetTable()
                              zoneFavorite = info.handler:GetSetting( moduleName, "profile", "ZoneFavorites" )[LibLordFarlander.UnnamifyString( info[#info - 1] )]
                          end--if
                          currentFavorite = zoneFavorite[ui.favorites[ui.currentFavorite]]
                          if( not currentFavorite ) then
                              zoneFavorite[ui.favorites[ui.currentFavorite]] = { frequency = val }
                              currentFavorite = zoneFavorite[ui.favorites[ui.currentFavorite]]
                          else
                              currentFavorite.frequency = val
                          end--if
                          if( not ( currentFavorite.frequency or currentFavorite.active or currentFavorite.exclusive ) ) then
                              zoneFavorite[ui.favorites[ui.currentFavorite]] = nil
                          end--if
                          info.handler:BuildZoneFavorites()
                          info.handler:ChooseRandomCompanion()
                      end,
            },
            active = {
                type = "toggle",
                order = 4,
                name = L["Override favorites"],
                desc = L["Check this to override the favorites settings for this companion."],
                get = function( info )
                          local zoneFavorite = info.handler:GetSetting( moduleName, "profile", "ZoneFavorites" )[LibLordFarlander.UnnamifyString( info[#info - 1] )]
                          local ui = info.handler:GetUISettings( moduleName )

                          if( zoneFavorite and zoneFavorite[ui.favorites[ui.currentFavorite]] ) then
                              return zoneFavorite[ui.favorites[ui.currentFavorite]].active or false
                          end--if
                          return false
                      end,
                set = function( info, val )
                          local zoneFavorite = info.handler:GetSetting( moduleName, "profile", "ZoneFavorites" )[LibLordFarlander.UnnamifyString( info[#info - 1] )]
                          local CurrentFavorite
                          local ui = info.handler:GetUISettings( moduleName )

                          if( not zoneFavorite ) then
                              info.handler:GetSetting( moduleName, "profile", "ZoneFavorites" )[LibLordFarlander.UnnamifyString( info[#info - 1] )] = TR:GetTable()
                              zoneFavorite = info.handler:GetSetting( moduleName, "profile", "ZoneFavorites" )[LibLordFarlander.UnnamifyString( info[#info - 1] )]
                          end--if
                          CurrentFavorite = zoneFavorite[ui.favorites[ui.currentFavorite]]
                          if( not CurrentFavorite ) then
                              zoneFavorite[ui.favorites[ui.currentFavorite]] = { active = val }
                              CurrentFavorite = zoneFavorite[ui.favorites[ui.currentFavorite]]
                          else
                              CurrentFavorite.active = val
                          end--if
                          if( not ( CurrentFavorite.frequency or CurrentFavorite.active or CurrentFavorite.exclusive ) ) then
                              zoneFavorite[ui.favorites[ui.currentFavorite]] = nil
                          end--if
                          info.handler:BuildZoneFavorites()
                          info.handler:ChooseRandomCompanion()
                      end,
            },
            exclusive = {
                type = "toggle",
                order = 5,
                name = L["Exclusive"],
                desc = L["Companions marked Exclusive will be the only ones picked from."],
                get = function( info )
                          local zoneFavorite = info.handler:GetSetting( moduleName, "profile", "ZoneFavorites" )[LibLordFarlander.UnnamifyString( info[#info - 1] )]
                          local ui = info.handler:GetUISettings( moduleName )

                          if( zoneFavorite and zoneFavorite[ui.favorites[ui.currentFavorite]] ) then
                              return zoneFavorite[ui.favorites[ui.currentFavorite]].exclusive or false
                          end--if
                          return false
                      end,
                set = function( info, val )
                          local zoneFavorite = info.handler:GetSetting( moduleName, "profile", "ZoneFavorites" )[LibLordFarlander.UnnamifyString( info[#info - 1] )]
                          local currentFavorite
                          local ui = info.handler:GetUISettings( moduleName )

                          if( not zoneFavorite ) then
                              info.handler:GetSetting( moduleName, "profile", "ZoneFavorites" )[LibLordFarlander.UnnamifyString( info[#info - 1] )] = TR:GetTable()
                              zoneFavorite = info.handler:GetSetting( moduleName, "profile", "ZoneFavorites" )[LibLordFarlander.UnnamifyString( info[#info - 1] )]
                          end--if
                          currentFavorite = zoneFavorite[ui.favorites[ui.currentFavorite]]
                          if( not currentFavorite ) then
                              zoneFavorite[ui.favorites[ui.currentFavorite]] = { exclusive = val }
                              currentFavorite = zoneFavorite[ui.favorites[ui.currentFavorite]]
                          else
                              currentFavorite.exclusive = val
                          end--if
                          if( not ( currentFavorite.frequency or currentFavorite.active or currentFavorite.exclusive ) ) then
                              zoneFavorite[ui.favorites[ui.currentFavorite]] = nil
                          end--if
                          info.handler:BuildZoneFavorites()
                          info.handler:ChooseRandomCompanion()
                      end,
            },
    }
    local continentTemplate = {
    }
    local zoneOptions = {
        type = "group",
        name = L["Zone Favorites"],
        desc = L["Set a companion to use when in a certain zone."],
        handler = MiniPet,
        childGroups = "tree",
        args = {
            [L["list"]] = {
                type = "execute",
                order = -1,
                guiHidden = true,
                name = L["List all favorite companions"],
                desc = L["Show a list with all of your favorite companions."],
                func = function( info )
                           local handler = info.handler
                           local Print = handler.Print
                           local lHowOften = L["%.1f times as often"]
                           local lExclusive = L["exclusive"]
                           local lNever = L["never"]
                           local GetItemLink = LibLordFarlander.GetItemLink
                           
                           Print( handler, L["List of your favorite companions:"] )
                           for zone, pets in pairs( handler:GetSetting( moduleName, "profile", "ZoneFavorites" ) ) do
                               Print( handler, ("%s:"):format( zone ) )
                               for pet, settings in pairs( pets ) do
                                   local howOften = settings.frequency or 1.0
                                   local sHowOften =  "default"
                                   local exclusive = settings.exclusive

                                   if( ( howOften ~= 1.0 ) or exclusive ) then
                                       local link = ( pet > 0 ) and GetItemLink( pet ) or GetSpellLink( pet * -1 )

                                       if( pet > 0 ) then
                                           sTexture = select( 10, GetItemInfo( pet ) )
                                       else
                                           sTexture = select( 3, GetSpellInfo( pet * -1 ) )
                                       end--if
                                       if( howOften == 0.0 ) then
                                           sHowOften = lNever
                                       elseif( howOften ~= 1.0 ) then
                                           sHowOften = lHowOften:format( howOften )
                                       end--if
                                       if( exclusive ) then
                                           sHowOften = ("%s, %s"):format( sHowOften, lExclusive )
                                       end--if
                                       if( sTexture ) then
                                           Print( handler, ("|T%s:0|t %s |cFFFF0000[%s]|r"):format( sTexture, link or tostring( pet ), sHowOften ) )
                                       else
                                           Print( handler, ("%s |cFFFF0000[%s]|r"):format( link or tostring( pet ), sHowOften ) )
                                       end--if
                                   end--if
                               end--if
                           end--if
                       end,
            },
        },
    }

    local CopyTable = TR.CopyTable
    local NamifyString = LibLordFarlander.NamifyString
    local GetMapZonesWithExtra  = LibLordFarlander.GetMapZonesWithExtra
    local ReturnTable = TR.ReturnTable
    
    for continentindex, continent in pairs( { GetMapContinents() } ) do
        zoneOptions.args[NamifyString( continent )] = {
            type = "group",
            handler = MiniPet,
            name = continent,
            desc = L["Zones in %s"]:format( continent ),
            childGroups = "tree",
            args = {},
        }

        local continentArgs = zoneOptions.args[NamifyString( continent )].args

        CopyTable( TR, continentTemplate, continentArgs )
        
        local mapZones = GetMapZonesWithExtra( continentindex )
        
        for index, zone in pairs( mapZones ) do
            continentArgs[NamifyString( zone )] = {
                type = "group",
                handler = MiniPet,
                name = zone,
                desc = zone,
                args = zoneTemplate,
            }
        end--for
        ReturnTable( TR, mapZones )
    end--for

	return zoneOptions
end--Companions:BuildZoneFavoritesOptions( MiniPet )

function Companions:OnBuildAdditionalOptions( MiniPet )
    local options = {
        [L["Favorites"]] = self:BuildFavoritesOptions( MiniPet ),
        [L["Zone Favorites"]] = self:BuildZoneFavoritesOptions( MiniPet ),
    }

    return options
end--Companions:OnBuildAdditionalOptions( MiniPet )

function Companions:OnCheckItem( MiniPet, itemID, bag, slot )
    return itemID and PetAndMountDatabase:ItemInSet( itemID, "companion" )
end--Companions:OnCheckItem( MiniPet, itemID, bag, slot )

function Companions:OnScanPlayer( MiniPet, dirtyBags, firstSeen )
    local returnVal = false

    if( dirtyBags[-255] ) then
        local critters = PetAndMountDatabase:GetPlayersCompanions( "CRITTER" )
        local spellList = self.playersCompanions.list[-255]
        local BuildSpell = ItemList.BuildSpell

        if( not spellList ) then
            self.playersCompanions.list[-255] = TR:GetTable()
            spellList = self.playersCompanions.list[-255]
        end--if
        for _, critter in pairs( critters ) do
		    local spellID = critter.spellID * -1

            if( not firstSeen[spellID] ) then
                firstSeen[spellID] = time()
            end--if
            if( not spellList[spellID] ) then
			    spellList[spellID] = BuildSpell( critter.spellID )
                returnVal = true
            end--if
        end--for
        TR:ReturnTable( critters )
    end--if
    return returnVal
end--Companions:OnScanPlayer( MiniPet, dirtyBags, playerPetList, firstSeen )

function Companions:OnStartBagScan( MiniPet, bag )
    if( self.playersCompanions.tempList ) then
        TR:ReturnTable( self.playersCompanions.tempList )
        self.playersCompanions.tempList = nil
    end--if
end--Companions:OnStartBagScan( MiniPet, bag )

function Companions:OnAddBagItem( MiniPet, bag, slot )
    local bagList = self.playersCompanions.tempList

    if( not bagList ) then
        self.playersCompanions.tempList = TR:GetTable()
        bagList = self.playersCompanions.tempList
    end--if
    ItemList.AddBagItemToList( bagList, bag, slot )
end--Companions:OnAddBagItem( MiniPet, bag, slot )

function Companions:OnAddInventoryItem( MiniPet, equipmentSlot )
    local bagList = self.playersCompanions.tempList

    if( not bagList ) then
        self.playersCompanions.tempList = TR:GetTable()
        bagList = self.playersCompanions.tempList
    end--if
    ItemList.AddInventoryItemToList( bagList, equipmentSlot )
end--Companions:OnAddInventoryItem( MiniPet, equipmentSlot )

function Companions:OnFinalizeBag( MiniPet, bag )
    local bagList = self.playersCompanions.tempList
    local playerList = self.playersCompanions.list
    local changed = false

    if( playerList[bag] ) then
        if( not TR.CompareTables( bagList, playerList[bag] ) ) then
            TR:ReturnTable( playerList[bag] )
            changed = true
            playerList[bag] = bagList
        else
            TR:ReturnTable( bagList )
        end--if
    elseif( bagList ) then
        changed = true
        playerList[bag] = bagList
    end--if
    self.playersCompanions.tempList = nil
    return changed
end--Companions:OnFinalizeBag( MiniPet, bag )

function Companions:OnCOMPANION_UPDATE( MiniPet, _, companionType )
    if( not companionType ) then
        MiniPet:DirtyBag( -255 )
    end--if
end--Companions:OnCOMPANION_UPDATE( MiniPet, _, companionType )

function Companions:OnPLAYER_REGEN_DISABLED( MiniPet )
    MiniPet:CombatStarted()
end--Companions:OnPLAYER_REGEN_DISABLED( MiniPet )

function Companions:OnPLAYER_REGEN_ENABLED( MiniPet )
    MiniPet:CombatEnded()
end--Companions:OnPLAYER_REGEN_ENABLED( MiniPet )

function Companions:OnZONE_CHANGED_NEW_AREA( MiniPet )
    MiniPet:GetUISettings( moduleName ).currentZone = GetRealZoneText()
    if( not TR.CompareTables( self.lastZoneList, self:OnGetZoneList() ) ) then
        MiniPet:ChooseRandomCompanion()
        self.lastZoneList = self:OnGetZoneList()
    end--if
    MiniPet:GetStateBasedOnModifiers()
end

function Companions:OnGetAllCompanionsList( MiniPet, indexed )
    local returnVal = TR:GetTable()

    for _, bag in pairs( self.playersCompanions.list ) do
        for _, item in pairs( bag ) do
            if( indexed ) then
                returnVal[item:GetID()] = item
            else
                tinsert( returnVal, item )
            end--if
        end--for
    end--for
    return returnVal
end--Companions:OnGetAllCompanionsList()

function Companions:OnHasCompanions( MiniPet )
    for _, bag in pairs( self.playersCompanions.list ) do
        for _, item in pairs( bag ) do
            return true
        end--for
    end--for
    return false
end--Companions:OnHasCompanions()

function Companions:OnBuildZoneFavorites( MiniPet )
    TR:ClearTable( self.playersCompanions.zones )
    self.playersCompanions.zones.default = TR:GetTable()

    local zones = self.playersCompanions.zones
    local companions = self:OnGetAllCompanionsList()
    local default = zones.default
    local profileFavorites = MiniPet:GetSetting( moduleName, "profile", "Favorites" )
    local GetTable = TR.GetTable
    local ReturnTable = TR.ReturnTable
    local HasTableMembers = LibLordFarlander.HasTableMembers

    for zone, list in pairs( MiniPet:GetSetting( moduleName, "profile", "ZoneFavorites" ) ) do
        local favorites = GetTable( TR )
        local exclusives = GetTable( TR )

        for pet, settings in pairs( list ) do
            if( settings.active ) then
                favorites[pet] = settings.frequency or 1.0
            end--if
            if( settings.exclusive ) then
                exclusives[pet] = settings.frequency or 1.0
            end--if
        end--if
        if( HasTableMembers( exclusives ) ) then
            ReturnTable( TR, favorites )
            zones[zone] = exclusives
        elseif( HasTableMembers( favorites ) ) then
            ReturnTable( TR, exclusives )
            for _, item in pairs( companions ) do
                local id = item:GetID()

                if( not favorites[id] ) then
                    favorites[id] = profileFavorites[id] or 1.0
                end--if
            end--if
            zones[zone] = favorites
        else
            ReturnTable( TR, exclusives )
            ReturnTable( TR, favorites )
        end--if
    end--for
    for  _, item in pairs( companions ) do
        local id = item:GetID()

        default[id] = profileFavorites[id] or 1.0
    end--if
    self.lastZoneList = self:OnGetZoneList()
    TR:ReturnTable( table.wipe( companions ) )
end--Companions:OnBuildZoneFavorites( MiniPet )

function Companions:OnGetZoneList()
    return self.playersCompanions.zones[GetRealZoneText()] or self.playersCompanions.zones.default
end--Companions:OnGetZoneList()

function Companions:OnSummonCompanion( MiniPet )
    if( self.currentCritterIndex and self.currentCompanion and IsUsableSpell( self.currentCompanion.spellID ) ) then
        MiniPet:Print( L["Summoning %s"]:format( select( 2, GetCompanionInfo( "CRITTER", self.currentCritterIndex ) ) ) )
        CallCompanion( "CRITTER", self.currentCritterIndex )
        MiniPet:ChooseRandomCompanion()
        MiniPet:SetButton( MiniPet:GetStateBasedOnModifiers() )
        return true
    end--if
    return nil
end--Companions:OnSummonCompanion()

function Companions:OnDismissCompanion()
    local summonedCritterIndex = PetAndMountDatabase:GetSummonedCompanionIndex( "CRITTER" )

    if( summonedCritterIndex ) then
        DismissCompanion( "CRITTER" )
    end--if
end--Companions:OnDismissCompanion()

function Companions:OnMouseClick( MiniPet, mouse, secure )
    if( ( mouse == "RightButton" ) and PetAndMountDatabase:GetSummonedCompanionIndex( "CRITTER" ) ) then
        MiniPet:SetButton( MiniPet:GetStateBasedOnModifiers() )
    end--if
    return nil
end--Companions:OnMouseClick( MiniPet, mouse, secure )

function Companions:OnChooseRandom( MiniPet )
    local companions = self:OnGetAllCompanionsList( true )

    self.currentCompanion = MiniPet:GetRandomFromListWithWeighing( companions, self:OnGetZoneList(), 0.0 )
    TR:ReturnTable( table.wipe( companions ) )
end--Companions:OnChooseRandom( MiniPet )

function Companions:OnGetCurrentCompanion( MiniPet )
    self:OnSetCurrentCompanion( MiniPet, self.currentCompanion )
    return self.currentCompanion
end--Companions:OnGetCurrentCompanion( MiniPet )

function Companions:OnSetCurrentCompanion( MiniPet, currentCompanion )
    if( currentCompanion and currentCompanion.spellID ) then
        self.currentCritterIndex = PetAndMountDatabase:FindCompanion( "CRITTER", currentCompanion.spellID )
    else
        self.currentCritterIndex = nil
    end--if
end--Companions:OnSetCurrentCompanion( MiniPet, currentCompanion )

function Companions:OnHasDismissableCompanionOut( MiniPet )
    local SummonedCritterIndex = PetAndMountDatabase:GetSummonedCompanionIndex( "CRITTER" )

    if( SummonedCritterIndex ) then
        return true
    end--if
    return nil
end--Companions:OnHasDismissableCompanionOut( MiniPet )

function Companions:OnIsCompanionSummonInCooldown( MiniPet )
    if( self.currentCritterIndex ) then
        local cooldown = GetCompanionCooldown( "CRITTER", self.currentCritterIndex )

        return cooldown > 0
    end--if
    return nil
end--Companions:OnIsCompanionSummonInCooldown( MiniPet )

function Companions:OnOutputWcpCode( MiniPet )
    local convertedCompanionList = {}
    local companionList = self:OnGetAllCompanionsList( MiniPet, false )

    for _, item in pairs( companionList ) do
        if( item:GetID() < 0 ) then
            tinsert( convertedCompanionList, item:GetID() * -1 )
        end--if
    end--for
    TR:ReturnTable( table.wipe( companionList ) )
    if( #convertedCompanionList ) then
        local petString = WarcraftPetsExpress:GenerateCodeSingle( convertedCompanionList )

        MiniPet:Print( L["Go to WarcraftPets.com and enter this code to update the pet list for this character:"] )
        MiniPet:Print( petString )
        ChatFrame_OpenChat( petString )
    else
        MiniPet:Print( L["Get a companion for some company!"] )
    end--if
end--Companions:OnOutputWcpCode( MiniPet )

LibStub( "LibLordFarlander-Randomizer-2.0" ):RegisterModuleToHandler( "MiniPet", moduleName, Companions )
