--[[
Copyright (c) 2010 WarcraftPets.com

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
]]--
--[[
  All variables are local three functions are provided to return a compressed string for pet IDs
  <string> = WarcraftPetsExpress:GenerateCodeSingle(tblSpellIDs); -- Collection is on single toon
  <string> = WarcraftPetsExpress:GenerateCodeMultiple(tblSpellIDs); -- Collection is on multiple toons
  <string> = WarcraftPetsExpress:GenerateCode(tblSpellIDs); -- Create string without setting collection type

  EXAMPLE
   local tblSpellIDs = {}; -- Build up a table of spell IDs
   table.insert(tblSpellIDs, 40990); -- Stinker
   table.insert(tblSpellIDs, 61991); -- Little Fawn
   table.insert(tblSpellIDs, 61773); -- Plump Turkey
   table.insert(tblSpellIDs, 55068); -- Mr. Chilly
   table.insert(tblSpellIDs, 65358); -- Calico Cat
   table.insert(tblSpellIDs, 10713); -- Albino Snake
   table.insert(tblSpellIDs, 67417); -- Obsidian Hatchling
   print(WarcraftPetsExpress:GenerateCodeSingle(tblSpellIDs)); -- Get code for NPC Breanni's pets this is the code that is copied into the clipboard
]]--
-- [Spell ID] = WarcraftPets.com ID, -- Name in WarcraftPets.com
local tblWarcraftPetsExpressPetDB = {
	[10675] = 1, -- Black Tabby Cat
	[10673] = 2, -- Bombay Cat
	[10674] = 3, -- Cornish Rex Cat
	[10676] = 4, -- Orange Tabby Cat
	[10677] = 5, -- Siamese Cat
	[10678] = 6, -- Silver Tabby Cat
	[10679] = 7, -- White Kitten
	[35239] = 8, -- Brown Rabbit
	[10711] = 9, -- Snowshoe Rabbit
	[4055] = 10, -- Mechanical Squirrel
	[10709] = 11, -- Brown Prairie Dog
	[28740] = 12, -- Whiskers the Rat
	[28739] = 13, -- Mr. Wiggles
	[17707] = 14, -- Panda Cub
	[15999] = 15, -- Worg Pup
	[10685] = 16, -- Ancona Chicken
	[12243] = 17, -- Mechanical Chicken
	[13548] = 18, -- Westfall Chicken
	[10707] = 19, -- Great Horned Owl
	[10706] = 20, -- Hawk Owl
	[10680] = 22, -- Cockatiel
	[10683] = 23, -- Green Wing Macaw
	[10682] = 24, -- Hyacinth Macaw
	[10684] = 25, -- Senegal
	[23811] = 26, -- Jubling
	[19772] = 27, -- Lifelike Toad
	[10704] = 28, -- Tree Frog
	[10703] = 29, -- Wood Frog
	[10714] = 30, -- Black Kingsnake
	[10716] = 31, -- Brown Snake
	[10717] = 32, -- Crimson Snake
	[28738] = 33, -- Speedy
	[10695] = 34, -- Dark Whelpling
	[10697] = 35, -- Crimson Whelpling
	[10698] = 36, -- Emerald Whelpling
	[32298] = 37, -- Netherwhelp
	[24696] = 38, -- Murky
	[27241] = 39, -- Gurky
	[24988] = 40, -- Lurky
	[36031] = 41, -- Blue Dragonhawk Hatchling
	[36027] = 42, -- Golden Dragonhawk Hatchling
	[36028] = 43, -- Red Dragonhawk Hatchling
	[36029] = 44, -- Silver Dragonhawk Hatchling
	[15067] = 45, -- Sprite Darter Hatchling
	[35907] = 46, -- Blue Moth
	[35909] = 47, -- Red Moth
	[35911] = 48, -- White Moth
	[35910] = 49, -- Yellow Moth
	[10688] = 50, -- Cockroach
	[35156] = 51, -- Mana Wyrmling
	[39181] = 52, -- Miniwing
	[33050] = 53, -- Magical Crawdad
	[25162] = 54, -- Disgusting Oozeling
	[16450] = 55, -- Smolderweb Hatchling
	[15049] = 56, -- Lil' Smoky
	[15048] = 57, -- Pet Bombling
	[26010] = 58, -- Tranquil Mechanical Yeti
	[26533] = 59, -- Father Winter's Helper
	[26541] = 60, -- Winter's Little Helper
	[26045] = 61, -- Tiny Snowman
	[26529] = 62, -- Winter Reindeer
	[27570] = 63, -- Peddlefeet
	[28871] = 64, -- Spirit of Summer
	[30156] = 65, -- Hippogryph Hatchling
	[17709] = 66, -- Zergling
	[17708] = 67, -- Mini Diablo
	[40549] = 68, -- Bananas
	[40614] = 69, -- Egbert
	[40634] = 70, -- Peanut
	[40613] = 71, -- Willy
	[40405] = 72, -- Lucky
	[36034] = 73, -- Firefly
	[39709] = 74, -- Wolpertinger
	[42609] = 75, -- Sinister Squashling
	[43918] = 76, -- Mojo
	[45082] = 77, -- Tiny Sporebat
	[10696] = 78, -- Azure Whelpling
	[45125] = 79, -- Rocket Chicken
	[45127] = 80, -- Dragon Kite
	[54187] = 92, -- Clockwork Rocket Bot
	[46599] = 99, -- Phoenix Hatchling
	[46426] = 100, -- Chuck
	[43698] = 101, -- Muckbreath
	[46425] = 102, -- Snarly
	[43697] = 103, -- Toothy
	[45890] = 104, -- Scorchling
	[49964] = 105, -- Ethereal Soul-Trader
	[51716] = 106, -- Nether Ray Fry
	[53082] = 108, -- Mini Tyrael
	[28505] = 109, -- Poley
	[23530] = 110, -- Tiny Red Dragon
	[23531] = 111, -- Tiny Green Dragon
	[45174] = 112, -- Golden Pig
	[45175] = 113, -- Silver Pig
	[48406] = 114, -- Spirit of Competition
	[48408] = 115, -- Essence of Competition
	[40990] = 117, -- Stinker
	[52615] = 118, -- Frosty
	[51851] = 119, -- Vampiric Batling
	[59250] = 120, -- Giant Sewer Rat
	[61472] = 121, -- Kirin Tor Familiar
	[53316] = 122, -- Ghostly Skull
	[61357] = 123, -- Pengu
	[61351] = 124, -- Cobra Hatchling
	[61350] = 125, -- Proto-Drake Whelp
	[61348] = 126, -- Tickbird Hatchling
	[61349] = 127, -- White Tickbird Hatchling
	[61855] = 128, -- Baby Blizzard Bear
	[61991] = 129, -- Little Fawn
	[62562] = 131, -- Ammen Vale Lashling
	[62508] = 132, -- Dun Morogh Cub
	[62513] = 133, -- Durotar Scorpion
	[62516] = 134, -- Elwynn Lamb
	[62564] = 135, -- Enchanted Broom
	[62491] = 136, -- Teldrassil Sproutling
	[62510] = 137, -- Tirisfal Batling
	[62674] = 138, -- Mechanopeep
	[62542] = 139, -- Mulgore Hatchling
	[63712] = 140, -- Sen'jin Fetish
	[62561] = 141, -- Strand Crawler
	[61725] = 142, -- Spring Rabbit
	[63318] = 143, -- Murkimus the Gladiator
	[62609] = 144, -- Argent Squire
	[62746] = 145, -- Argent Gruntling
	[65381] = 146, -- Curious Oracle Hatchling
	[65382] = 152, -- Curious Wolvar Pup
	[65682] = 153, -- Warbot
	[66520] = 154, -- Jade Tiger
	[66096] = 155, -- Shimmering Wyrmling
	[67417] = 156, -- Obsidian Hatchling
	[67413] = 157, -- Darting Hatchling
	[67414] = 158, -- Deviate Hatchling
	[67415] = 159, -- Gundrak Hatchling
	[67416] = 160, -- Leaping Hatchling
	[67420] = 161, -- Razzashi Hatchling
	[67418] = 162, -- Ravasaur Hatchling
	[67419] = 163, -- Razormaw Hatchling
	[66030] = 164, -- Grunty
	[44369] = 165, -- Pint-Sized Pink Pachyderm
	[55068] = 166, -- Mr. Chilly
	[65358] = 167, -- Calico Cat
	[10713] = 168, -- Albino Snake
	[68767] = 169, -- Tuskarr Kite
	[68810] = 170, -- Spectral Tiger Cub
	[61773] = 171, -- Plump Turkey
	[69002] = 172, -- Onyxian Whelpling
	[69541] = 173, -- Pandaren Monk
	[69677] = 174, -- Lil' K.T.
	[70613] = 175, -- Perky Pug
	[69452] = 176, -- Core Hound Pup
	[71840] = 177, -- Toxic Wasteling
	[69536] = 178, -- Wind Rider Cub
	[69535] = 179, -- Gryphon Hatchling
	[1] = 180, -- Purple Jade Leopard (Spell ID needed)
	[75134] = 181, -- Blue Clockwork Rocket Bot
	[75906] = 182, -- Lil&#39; XT
}
-- Takes two inputs between 0 and 511 and joins them into three characters low byte A, low byte B, high byte A & B
local WarcraftPetsExpress_strConvert = "qwertyuiopasdfghjklmnbvcxzQWERTYUIOPLKJHGFDSAZXCVBNM()1234567890";
local function WarcraftPetsExpress_EncodeNumbers(intA,intB)
	local pos0 = mod(intA,string.len(WarcraftPetsExpress_strConvert))+1;
	local pos1 = mod(intB,string.len(WarcraftPetsExpress_strConvert))+1;
	local pos2 = floor(intA/string.len(WarcraftPetsExpress_strConvert))*sqrt(string.len(WarcraftPetsExpress_strConvert))+floor(intB/string.len(WarcraftPetsExpress_strConvert))+1;
	local str0 = string.sub(WarcraftPetsExpress_strConvert,pos0,pos0);
	local str1 = string.sub(WarcraftPetsExpress_strConvert,pos1,pos1);
	local str2 = string.sub(WarcraftPetsExpress_strConvert,pos2,pos2);
	return str0..str1..str2;
end

-- Loops through a table of SpellIDs gets the WarcraftPets.com ID and encodes it, unknown pets are appended in the form =<spellid>=<petname> any non a-z in petname are removed as they are only to give hints for adding to the database
local function WarcraftPetsExpress_GenerateCode(tblSpellIDs)
	local strString = "";
	local strMissing = "";
	local intLastCode = 0;
	local intTotal = 0;
	local strKey = "";
	local strValue = "";
	for strKey,strValue in pairs(tblSpellIDs) do
		if tblWarcraftPetsExpressPetDB[strValue] then
			intTotal = intTotal + tblWarcraftPetsExpressPetDB[strValue];
			if intLastCode==0 then
				intLastCode = tblWarcraftPetsExpressPetDB[strValue];
			else
				strString = strString..WarcraftPetsExpress_EncodeNumbers(intLastCode,tblWarcraftPetsExpressPetDB[strValue]);
				intLastCode = 0;
			end
		else
			local strSpellName = GetSpellInfo(strValue);
			strSpellName = string.gsub(strSpellName,"[^a-zA-Z]","");
			strMissing = strMissing.."="..strValue.."="..strSpellName;
		end;
	end;
	if intLastCode~=0 then
		strString = strString..WarcraftPetsExpress_EncodeNumbers(intLastCode,0);
	end
	strString = strString..string.sub(WarcraftPetsExpress_strConvert,mod(intTotal,string.len(WarcraftPetsExpress_strConvert))+1,mod(intTotal,string.len(WarcraftPetsExpress_strConvert))+1);
	strString = strString..strMissing;
	return strString;
end

-- Global Functions 0 = Single Toon Collection, 1 = Multiple Toon Collection, 2 = A Collection, open for expansion later
WarcraftPetsExpress = {};

function WarcraftPetsExpress:GenerateCodeSingle(tblSpellIDs)
	return "0"..WarcraftPetsExpress_GenerateCode(tblSpellIDs);
end

function WarcraftPetsExpress:GenerateCodeMultiple(tblSpellIDs)
	return "1"..WarcraftPetsExpress_GenerateCode(tblSpellIDs);
end

function WarcraftPetsExpress:GenerateCode(tblSpellIDs)
	return "2"..WarcraftPetsExpress_GenerateCode(tblSpellIDs);
end
